
module Ansi 
  ( 
    -- ANSI color magic 
    Color(..) , sgr
  , withCol , withBgnd , withColors
  , putStrCol  , putStrLnCol
  , putStrCol' , putStrLnCol'
  , clrscr , topleft
  
    -- debug messages
  , submissionMode , whenSubmission, whenHuman, humanMessage, humanPrint
  , whenDebug, debugMsg, debugPrint
  
    -- ANSI input
    -- unfortunately, ANSI input is in Mine.hs since the data types are there
    -- and circular imports are not allowed
--  , getInteractiveMove , testInteractiveMove
--  , getLeftRightKey

  ) 
  where
 
--------------------------------------------------------------------------------

import Data.Char

import Data.IORef
import System.IO.Unsafe
import Control.Monad

--------------------------------------------------------------------------------

debugMsg :: String -> IO ()
debugMsg x = whenDebug $ putStrLn x

debugPrint :: Show a => a-> IO ()
debugPrint x = debugMsg (show x)

whenDebug :: IO () -> IO ()
-- whenDebug = id
whenDebug _ = return ()

--------------------------------------------------------------------------------

submissionMode :: IORef Bool
submissionMode = unsafePerformIO $ newIORef True

whenSubmission :: IO () -> IO ()
whenSubmission action = do
  b <- readIORef submissionMode
  when b $ action

whenHuman :: IO () -> IO ()
whenHuman action = do
  b <- readIORef submissionMode
  unless b $ action

humanMessage :: String -> IO ()
humanMessage xx = whenHuman $ putStrLn xx

humanPrint :: Show a => a -> IO ()
humanPrint x = whenHuman $ print x

--------------------------------------------------------------------------------

--ansiNum  :: [Int]  -> String
--ansiNum xs = "\ESC[" ++ map chr xs

ansiChar :: Char -> String
ansiChar x = "\ESC[" ++ [x]

ansiChars :: [Char] -> String
ansiChars xs = "\ESC[" ++ xs

sgr k = ansiChars ( show k ++ "m" )

--------------------------------------------------------------------------------

clrscr :: IO ()
clrscr = do
  putStr $ ansiChars "2J"

topleft :: IO ()
topleft = do
  putStr $ ansiChars "1;1H"

--------------------------------------------------------------------------------

data Color
  = Black
  | Red
  | Green
  | Yellow
  | Blue
  | Magenta
  | Cyan
  | White
  deriving (Eq,Show,Enum)

--------------------------------------------------------------------------------

withCol :: Color -> IO a -> IO a
withCol col action = do
  putStr $ sgr (30+fromEnum col) 
  x <- action
  putStr $ sgr 0 
  return x
  
withBgnd :: Color -> IO a -> IO a
withBgnd col action = do
  putStr $ sgr (40+fromEnum col) 
  x <- action
  putStr $ sgr 0 
  return x
  
withColors :: (Color,Color) -> IO a -> IO a
withColors (fore,back) action = do
  putStr $ sgr (30+fromEnum fore) 
  putStr $ sgr (40+fromEnum back) 
  x <- action
  putStr $ sgr 0 
  return x
  
--------------------------------------------------------------------------------
  
putStrCol :: Color -> String -> IO ()
putStrCol fore str = putStr $ concat [ sgr (30+fromEnum fore) , str , sgr 0 ]

putStrLnCol :: Color -> String -> IO ()
putStrLnCol fore str = putStrLn $ concat [ sgr (30+fromEnum fore) , str , sgr 0 ]

putStrCol' :: (Color,Color) -> String -> IO ()
putStrCol' (fore,back) str = putStr $ concat [ sgr (30+fromEnum fore) , sgr (40+fromEnum back) , str , sgr 0 ]

putStrLnCol' :: (Color,Color) -> String -> IO ()
putStrLnCol' (fore,back) str = putStrLn $ concat [ sgr (30+fromEnum fore) , sgr (40+fromEnum back) , str , sgr 0 ]

--------------------------------------------------------------------------------

