


-- we could use intmaps/intsets for pos, since x/y coords are very likely less than 2^15

{-# LANGUAGE BangPatterns #-}
module PosSet where

--------------------------------------------------------------------------------

import Data.Array
import Data.Array.IO
import Data.Bits
import Data.Ix

import Data.IntSet (IntSet) ; import qualified Data.IntSet as ISet
import Data.IntMap (IntMap) ; import qualified Data.IntMap as IMap

--------------------------------------------------------------------------------

newtype Pos = Pos Int deriving (Eq,Ord) -- ,Show)

instance Show Pos where show p = "{" ++ show x ++ "," ++ show y ++ "}" where (x,y) = unPos p

{-# INLINE mkPos #-}
mkPos :: Int -> Int -> Pos
mkPos x y = Pos $ posIndex (OldPos x y)

{-# INLINE unPos #-}
unPos :: Pos -> (Int,Int)
unPos (Pos k) = let OldPos x y = indexPos k in (x,y)

{-# INLINE unPos' #-}
unPos' :: Pos -> OldPos
unPos' (Pos k) = indexPos k 

{-# INLINE dePos #-}
dePos :: Pos -> Int
dePos (Pos k) = k

--------------------------------------------------------------------------------

getOldBounds :: IOArray Pos a -> IO (OldPos,OldPos)
getOldBounds mine = do
  (p1,p2) <- getBounds mine
  let (a,b) = unPos p1
      (n,m) = unPos p2
  return (OldPos a b, OldPos n m) 

oldBounds :: Array Pos a -> (OldPos,OldPos)
oldBounds mine = (OldPos a b, OldPos n m) where
  (p1,p2) = bounds mine
  (a,b) = unPos p1
  (n,m) = unPos p2

--------------------------------------------------------------------------------

instance Ix Pos where
  
  {-# INLINE range #-}
  range (p1,p2) = [ mkPos x y | x<-[x1..x2], y<-[y1..y2] ] where 
    (x1,y1) = unPos p1
    (x2,y2) = unPos p2

  {-# INLINE index     #-}
  {-# INLINE inRange   #-}
  {-# INLINE rangeSize #-}
  index     (p1,p2) q = index     (unPos p1, unPos p2) (unPos q)
  inRange   (p1,p2) q = inRange   (unPos p1, unPos p2) (unPos q)   
  rangeSize (p1,p2)   = rangeSize (unPos p1, unPos p2) 
  
--------------------------------------------------------------------------------

newtype PSet   = PSet { unPSet :: IntSet   }
newtype PMap a = PMap { unPMap :: IntMap a }

data OldPos = OldPos {-# UNPACK #-} !Int {-# UNPACK #-} !Int deriving (Eq,Ord,Show,Ix)

{-# INLINE posIndex #-}
posIndex :: OldPos -> Int
posIndex (OldPos !x !y) = {- 2147483647 -   -}  ( x + (shiftL y 15) )
-- minus sign is opposite order; y is high bit so first priority

{-# INLINE indexPos #-}
indexPos :: Int -> OldPos
--indexPos k = let n = (2147483647-k) in OldPos (n .&. 32767) (shiftR n 15)
indexPos !n = OldPos (n .&. 32767) (shiftR n 15)

--------------------------------------------------------------------------------

setEmpty :: PSet
setEmpty = PSet (ISet.empty)

setSingleton :: Pos -> PSet
setSingleton (Pos pos) = PSet (ISet.singleton pos)

setUnion, setDifference :: PSet -> PSet -> PSet
setUnion      (PSet s) (PSet t) = PSet (ISet.union      s t)
setDifference (PSet s) (PSet t) = PSet (ISet.difference s t)

setInsert, setDelete :: Pos -> PSet -> PSet 
setInsert (Pos pos) (PSet s) = PSet (ISet.insert pos s)
setDelete (Pos pos) (PSet s) = PSet (ISet.delete pos s)
--setInsert pos (PSet s) = PSet (ISet.insert (posIndex pos) s)
--setDelete pos (PSet s) = PSet (ISet.delete (posIndex pos) s)

setFromList :: [Pos] -> PSet
setFromList = PSet . ISet.fromList . map dePos {- posIndex -}

setToList :: PSet -> [Pos] 
setToList = map Pos {- indexPos -} . ISet.toList . unPSet

--------------------------------------------------------------------------------

{-
-- we invert the order, so that there is no need to reverse a list
-- instance Ord Pos where compare (Pos x1 y1) (Pos x2 y2) = compare (y1,x1) (y2,x2)
instance Ord Pos where compare (Pos x1 y1) (Pos x2 y2) = compare (y2,x2) (y1,x1) 
-}

