
{- 

the first plan is to do a pruned brute-force search

the cost function is the points gained (say 33*lambda-moves) plus an estimation
of the future costs. at zeroth level we can estimate the future cost by
computing the distances of the remaining lambdas from us.

at a more sophisticated level, we could start similar pruned searches with
much more agressive pruning?

-}

module Search where

--------------------------------------------------------------------------------

import Data.Array
import Data.Array.IO
import Data.Array.MArray

import Data.List
import Data.Maybe

import Control.Monad
import Control.Applicative

import Data.Char
import System.IO
import Data.IORef

import Data.Map (Map) ; import qualified Data.Map as Map
import Data.Set (Set) ; import qualified Data.Set as Set

import Ansi
import PosSet
import Spatial
import Mine
import Simulation
import SimBetter
import AStar

--------------------------------------------------------------------------------

calculateStateScore :: SimState -> Double
calculateStateScore = computeStatus

computeStatus :: SimState -> Status
computeStatus state = 
  score
{-
  Status { _statScore = score }
-}
  where
    score = fromIntegral scoreDone 
          + 0.70 * scoreEstimated 
          + scoreEndstate 
          + 0.80 * (_simSumMMScore state)     -- sum of movement scores
          - fromIntegral (_simBeardCnt state)
          
    -- what already happened
    scoreDone = 66 * (_simLambdaCnt state) - (_simMoveCnt state)
    
    scoreEndstate = case stateEndCond state of
      Nothing -> 0
      Just ec -> case ec of
        Winning     ->  1000000
        Aborted     -> -20
        Losing      -> -666666
        Inoperative -> -666666
    
{- 
-- this does not seems to be a good idea
    histpos = map fst (_simRevHistory state) :: [Pos]
    repeatPenalty = sum $ map (\x -> (x-1)) $ map length $ group $ sort histpos
-}

    -- remaining = bsetToUnsortedList (_simLambdaSet state)
    nremaining = (_simTotalLambda state) - (_simLambdaCnt state)     -- now with higher order rocks these are different!

    liftpos = _simLift     state
    ourpos  = _simRobotPos state
    -- horocks = bsetToUnsortedList (_simHorockSet state)
    
    near_horocks   = {-# SCC "near_horocks" #-} bsetFindClose 15 ourpos (_simHorockSet state) -- horocks
    near_remaining = {-# SCC "near_remain"  #-} bsetFindClose 20 ourpos (_simLambdaSet state) -- remaining
    
    -- I have no idea of estimating a triple distance cost is pessimistic enough...
    -- in any case this is a very stupid cost function

    scoreEstimated = if (nremaining == 0)
      then - 0.1 * superDist ourpos liftpos 
      else - 0.1 * superDist ourpos liftpos + 
           sum' [ max 0 (50 - 2   * superDist ourpos lampos) | lampos <- near_remaining ] +
           sum' [ max 0 (30 - 1.2 * superDist ourpos horpos) | horpos <- near_horocks   ]
            
--    scoreEstimated = if (nremaining == 0)
--      then 10000 - 3 * superDist ourpos liftpos 
--      else 10 * fromIntegral near_nremaining
           
superDist p q 
  = 0.3333 * manhattanDistf   p q
  + 0.3333 * euclideanDist    p q
  + 0.3333 * sqrt (euclideanDist p q)
--  + 0.1111 * euclideanDistSqr p q    -- stupid stupid, negyzetosszeg kozepen minimalis :(
  
computeEndstateScore :: Result -> Double
computeEndstateScore result = fromIntegral score    
  where
    score   = resultPoints result - penalty 
    penalty = (if losingResult result then 666666 else 0)     -- losing is the work of satan!
    
--------------------------------------------------------------------------------

data SearchCfg = SearchCfg
  { _cfgDepth     :: Int     -- ^ how long we look into the future
  , _cfgSurvivors :: Int     -- ^ how many survivor paths we allow in each step
  }
  deriving Show

-- what we should do for pruned search is probably a search trie!


--------------------------------------------------------------------------------

depthFirstSearchAll :: Int -> SimState -> IO [Path]
depthFirstSearchAll = naiveDepthFirstSearch

-- ok this turned out to be a depth-first search instead of a pruned breadth-first...
-- actually depth-first works much more elegantly, so maybe i will use this :)
naiveDepthFirstSearch :: Int -> SimState -> IO [Path]
naiveDepthFirstSearch depth ini_simstate = 
  do
    collector <- newIORef [] :: IO (IORef [Path])      -- we collect final paths here 
    search collector ini_simstate  
    readIORef collector
  where
    search collector simstate0__ = go 0 emptyPath simstate0__ where
      go len path@(Path unpath) simstate = if len >= depth
        then modifyIORef collector (path:) 
        else do
          -- ok, here comes the meat (but i'm vegetarian?? in theory at least)
          let meta = _simMeta     simstate
              mine = _simMine     simstate
              pos  = _simRobotPos simstate
          moves <- generateMoves meta mine pos        
          forM_ moves $ \move -> do

            simulateMoveWithUndo simstate move $ \st -> case eiSimState st of

              Right simstate' -> do
                let status = computeStatus simstate'
                    changes = _simLastChanges simstate'
                    step = Step move status changes
                -- now we recurse
                go (len+1) (Path (step:unpath)) simstate'

              Left (result,simstate') -> do
                let endScore = computeEndstateScore result 
                    status0  = computeStatus simstate'
                    status   = status0 + endScore -- status0 { _statScore = endScore + (_statScore status0) }
                    changes = _simLastChanges simstate'
                    step = Step move status changes
                    thispath = Path (step:unpath)
          --      print status
                modifyIORef collector (thispath:)
                                
                
--------------------------------------------------------------------------------
-- brute-force depth first search 

naiveNextMoves :: Int -> SimState -> IO [Move]
naiveNextMoves = depthFirstSearchBest

depthFirstSearchBest :: Int -> SimState -> IO [Move]
depthFirstSearchBest depth state = do
  -- tmpstate <- cloneState state               -- better safe than sorry... NAH, LET'S PLAY DANGEREOUSLY!
  paths <- naiveDepthFirstSearch depth state    -- tmpstate
  return $ bestPathMoves paths

--------------------------------------------------------------------------------
-- breadth first search with pruning
        
prunedNextKSteps :: Int -> Int -> SimState -> IO [Path]        
prunedNextKSteps k nsurvivors state = do
  let initialScore = computeStatus state

  paths <- naiveDepthFirstSearch k state
  let sortedpaths = groupBy (equating pathScore) $ sortBy (reverseComparing pathScore) paths
--   print $ (map.map) pathScore $ take 40 sortedpaths
  return {- $ filter (\p -> pathScore p > initialScore - 100) -} $ map head $ take nsurvivors sortedpaths

prunedPaths :: Int -> Int -> SimState -> IO [Path]
prunedPaths depth nsurvivors = go 0 [emptyPath] where
  go k paths state = if (k>=depth) 
    then return paths
    else do
      fullpaths <- forM paths $ \path@(Path unpath) -> do
        simulatePathWithUndo state path $ \st -> case eiSimState st of
          Left {}      -> return [path]
          Right state' -> do
--            print (_simRobotPos state')  ----------
            paths' <- prunedNextKSteps 3 nsurvivors state'
--            print paths'
            return $ if length unpath >= 6   -- 2 lepes utan engedjuk megallni, ha nagyon akar  
              then (Path unpath) : [ Path (new ++ unpath) | Path new <- paths' ] 
              else                 [ Path (new ++ unpath) | Path new <- paths' ]                    
      let sortedpaths = sortBy (reverseComparing pathScore) (concat fullpaths)
      let newpaths = take nsurvivors sortedpaths
      go (k+1) newpaths state
      
prunedNextMoves :: Int -> Int -> SimState -> IO [Move]
prunedNextMoves depth4 nsurvivors state = do
  -- tmpstate <- cloneState state       -- better safe than sorry... NAH, LIVE FAST
  paths    <- prunedPaths depth4 nsurvivors state -- tmpstate
  return $ bestPathMoves paths

bestPathMoves :: [Path] -> [Move]
bestPathMoves paths = moves where
  sortedpaths = sortBy (reverseComparing pathScore) paths
  moves = case sortedpaths of
    [] -> [Abort]
    _  -> case steps of
            [] -> [Wait]         -- shouldn't normally happen
            _  -> reverse $ map _stepMove steps 
          where
            bestpaths@(Path steps) = head sortedpaths 

bestPathMove :: [Path] -> Move
bestPathMove paths = head $ bestPathMoves paths
     
--------------------------------------------------------------------------------

type NextMoves = SimState -> IO [Move]

strategy :: Meta -> Mine -> IO [Move]
strategy meta puremine = 
  do
    inistate  <- initialSimState meta puremine
    collector <- newIORef [] :: IO (IORef [Move]) 
    doit collector inistate
    reverse <$> readIORef collector
  where
    maxStepsAllowed = extendedMineSize puremine - 1
    doit collector inistate = go inistate where
      go state = do
        move <- if ( _simMoveCnt state > maxStepsAllowed )
          then return Abort
          else -- head <$> prunedNextMoves 10 20 state 
               head <$> naiveNextMoves 7 state
        modifyIORef collector (move:)
--        print (_simRevHistory state)
        print move
        st <- simulateStep state move  
        case eiSimState st of
          Right state' -> go state'
          Left  (result,state') -> print result

--------------------------------------------------------------------------------
-- we will actually execute the plans, whatever it costs!
    
determinedStrategy :: Meta -> Mine -> IO [Move]    
determinedStrategy meta puremine = 
  do
    inistate  <- initialSimState meta puremine
    collector <- newIORef [] :: IO (IORef [Move]) 
    doit collector inistate
    reverse <$> readIORef collector
  where
    maxStepsAllowed = extendedMineSize puremine - 1
    doit collector inistate = go [] inistate where
    
      go plan@(this:rest) state = makeMove this state (go rest)
      go [] state = do
        moves <- if ( _simMoveCnt state > maxStepsAllowed )
          then return [Abort]
          else prunedNextMoves 30 500 state 
               -- naiveNextMoves 7 state               
--        print (_simRevHistory state)
        -- print (_simRobotPos state , moves)
{-
        case take 6 moves of
          [Move L,Move R,Move L,Move R,Move L,Move R] -> go [Abort] state
          [Move R,Move L,Move R,Move L,Move R,Move L] -> go [Abort] state
          [Move U,Move D,Move U,Move D,Move U,Move D] -> go [Abort] state
          [Move D,Move U,Move D,Move U,Move D,Move U] -> go [Abort] state
          _ -> go (take 2 moves) state
-}
        go (take 5 moves) state

      -- i'm using continuation, i must be drunk!
      makeMove move state cont = do
        modifyIORef collector (move:)
        humanPrint move
        whenSubmission $ putChar (encodeMove move)
        st <- simulateStep state move  
        case eiSimState st of
          Right state' -> cont state'
          Left  (result,state') -> humanPrint result         
           
--------------------------------------------------------------------------------
      
the_old_strategy = determinedStrategy
-- the_strategy = determinedStrategy

      
--------------------------------------------------------------------------------
  
            