
-- | less naive simulator (we only update the positions watched)

module SimBetter where

--------------------------------------------------------------------------------

import Data.Array
import Data.Array.IO
import Data.Array.MArray

import Data.List
import Data.Maybe

import Control.Monad
import Control.Applicative

import Data.IORef
import System.IO

import Data.Map (Map) ; import qualified Data.Map as Map
import Data.Set (Set) ; import qualified Data.Set as Set

import PosSet
import Spatial
import Mine
import Simulation
import Ansi
    
--------------------------------------------------------------------------------

simulateMoveWithUndo :: SimState -> Move -> (SimState -> IO a) -> IO a
simulateMoveWithUndo = simulateMoveWithUndo' simulateStep

simulateMovesWithUndo :: SimState -> [Move] -> (SimState -> IO a) -> IO a
simulateMovesWithUndo = simulateMovesWithUndo' simulateStep

simulatePathWithUndo :: SimState -> Path -> (SimState -> IO a) -> IO a
simulatePathWithUndo = simulatePathWithUndo' simulateStep 

--------------------------------------------------------------------------------

updateExtendedMine' :: BSet -> (Pos,Bool,Bool) -> Watch -> IOMine -> IO ([Write],[Undo])
updateExtendedMine' beardset (liftpos,beardFlag,readyFlag) watchSet mine = do
  (OldPos a b, OldPos n m) <- getOldBounds mine

  collector <- newIORef [] :: IO (IORef [[Write]])

  -- hackety hack hack
  watchSet' <- if not beardFlag
    then return watchSet 
    else do
      -- bpos <- ioFindBeards mine 
      -- return $ Set.union watchSet (Set.fromList bpos)
      return $ setUnion watchSet (bsetToPset beardset)

  -- FORTUNATELY, Set.toList gives an ascending list, and the Ord instance
  -- of Pos is defined so that it is exactly the update order
  -- 
  -- except that because of the collection procedure, the order is reversed...
  -- SO, INVERT the order!
  
  forM_ (setToList watchSet') $ \pos -> do  
    ws <- updateCell' beardFlag readyFlag mine pos  
    modifyIORef collector (ws:)    
   
  -- somewhat ugly but who cares :)
  when readyFlag $ do
    ws <- updateCell' beardFlag readyFlag mine liftpos   
    modifyIORef collector (ws:)    
   
--  allwrites <- (concat {- . reverse -} ) <$> readIORef collector
  allwrites <- (concat . reverse ) <$> readIORef collector
  allundos  <- executeWritesWithUndo mine allwrites
  
  return (allwrites,allundos)
  
--------------------------------------------------------------------------------

simulateStepEi :: SimState -> Move -> IO (Either (Result,SimState) SimState)
simulateStepEi oldstate move = do
  state' <- simulateStep oldstate move 
  return $ if stateIsFinished state'
    then Left (fromJust (_simResult state'), state')
    else Right state'
    
simulateStep :: SimState -> Move -> IO SimState -- (Either (Result,SimState) SimState)
simulateStep oldstate move = do

  let SimState pos liftpos lamcnt lamset beardset trampmap horockset totallam totlamnorock movcnt 
               nrazors beardcnt waterlevel underwater revhist watch changeset oldmmm mmscores simresult mine meta = oldstate

  let revhist' = (pos,move) : revhist

  let movcnt' = movcnt + 1     -- according to validator, each step counts

  let flooding    = _metaFlooding    meta 
      trampolines = _metaTrampolines meta    
      hair        = _metaHair        meta
      growth      = _growth hair 

  let beardFlag = mod movcnt' growth == 0
      
  let waterlevel' = (_waterLevel flooding) + water_incr where
        water_incr = case (_floodRate flooding) of 
          Nothing   -> 0
          Just rate -> div movcnt' rate  
            
  -- "mmm" = Maybe MoveMent
  mmm <- canMove oldstate {- meta mine -} pos move
  let (pos',lamcnt',lamset',nrazors',writes1) = case mmm of
        Nothing -> (pos,lamcnt,lamset,nrazors, [])
        Just mm -> case mm of
                     LambdaCollected {} -> (pos', lamcnt+1, bsetDelete pos' lamset, nrazors  , writes1)                     
                     RazorCollected  {} -> (pos', lamcnt  ,                 lamset, nrazors+1, writes1)
                     Shaved          {} -> (pos', lamcnt  ,                 lamset, nrazors-1, writes1) 
                     _                  -> (pos', lamcnt  ,                 lamset, nrazors  , writes1)  
                   where
                     (pos',writes1) = writeMovement meta mm 
                     
  undos1 <- executeWritesWithUndo mine writes1
    
  let actionWatch = writeWatch writes1            -- we have to check to new move
      updateWatch = setUnion watch actionWatch    -- and what was changed in the previous update
  
  let underwater' = if ycoord pos' <= waterlevel'
        then underwater + 1
        else 0

  let mmscores' = mmscores + mmovementScore mmm 
                + fromIntegral (underwater' - underwater) * 5    -- levegovetel pont

-- ennek nem itt a helye hanem update-ben:  when (movcnt' mod growth == 0) $ do       

  let trampmap'  = foldr Map.delete trampmap           [ pos | Write pos Empty       <- writes1 ]     -- !!  
        
  let horockset1 = case mmm of 
        Just (MoveAndPush pos dir True) -> bsetInsert pos'' (bsetDelete pos' horockset) where
          pos'  = moveByDir pos  dir
          pos'' = moveByDir pos' dir
        _ -> horockset
              
--  let beardset1 = (setDifference beardset wasshaved) where
--        wasshaved = case mmm of { Just (Shaved pos these) -> setFromList these ; _ -> setEmpty }  
  let (beardcnt1,beardset1) = case mmm of 
        Just (Shaved pos these) -> (beardcnt - length these, bsetDifference1 beardset these) -- (setFromList these) 
        _ -> (beardcnt,beardset) 
        
        
  (writes,undos) <- updateExtendedMine' beardset1 (liftpos , beardFlag , lamcnt' >= totallam) updateWatch mine            
  loser  <- checkLosingCond pos' writes mine
  
  let (beardcnt',beardset') = if not beardFlag 
        then (beardcnt1 , beardset1) 
        else (beardcnt1 + length newgrowth , beardset1 `bsetUnion1` newgrowth) where -- (setFromList newgrowth) where
               newgrowth = [ pos | Write pos Beard <- writes ]
  
  let horockset' = (\middle -> foldr bsetInsert middle [ pos | Write pos (Rock True) <- writes ])
                 $ foldr bsetDelete horockset1         [ pos | Write pos Empty       <- writes ] 
  
  let watch' = writeWatch writes     -- we don't need the action here ??? (i think so...)
  
  let changeset' = ChangeSet (writes1++writes) (undos++undos1)   -- undos in opposite order of course
  
  let tmpResult = Result 
        { _endCond   = error "end condition"
        , _lambdaCnt = lamcnt'
        , _moveCnt   = movcnt'
        , _history   = History (map snd $ reverse revhist')
        }

  let newstate = oldstate 
        { _simRobotPos    = pos' 
        , _simLambdaCnt   = lamcnt' 
        , _simLambdaSet   = lamset'
        , _simBeardSet    = beardset'
        , _simTrampolins  = trampmap'
        , _simHorockSet   = horockset'
        , _simMoveCnt     = movcnt'
        , _simRazors      = nrazors'
        , _simBeardCnt    = beardcnt'
        , _simRevHistory  = revhist'
        , _simWaterLevel  = waterlevel'
        , _simUnderWater  = underwater'
        , _simWatch       = watch'
        , _simLastChanges = changeset'
        , _simLastMovement = mmm
        , _simSumMMScore  = mmscores'
        }

  let -- finished x = return $ Left (x,newstate) 
      finished res = return $ newstate { _simResult = Just res }
        
  -- Order is the following:
  --  1) check for win
  --  2) abort (updated specification)
  --  3) losing
  case mmm of      
    Just (MineCompleted {}) -> finished $ tmpResult { _endCond = Winning }            
    _ -> case move of
      Abort -> finished $ tmpResult { _endCond = Aborted }     
      _ -> if loser
        then finished $ tmpResult { _endCond = Losing }
        else if underwater' > _waterProof flooding 
          then finished $ tmpResult { _endCond = Inoperative } 
          else return $ {-Right-} newstate

--------------------------------------------------------------------------------

interactiveSim :: Meta -> Mine -> IO ()
interactiveSim = interactiveSimWith simulateStep

--------------------------------------------------------------------------------

main_sim fn = do
  (meta,mine) <- loadMap fn -- "../maps/flood1.map" -- contest2.map" -- flood5.map"
  print meta
  interactiveSim meta mine
  
main_play fn movestr = do
  let moves = parseMoves movestr  
  (meta,mine) <- loadMap fn 
  print meta
  timeMachine meta mine moves

timeMachine meta mine moves = do
  zipper <- zipperSimulationWith simulateStep meta mine moves 
  zipperPlaybackWith simulateStep meta mine zipper 

main_dump fn movestr = do
  let moves = parseMoves movestr  
  (meta,mine) <- loadMap fn 
  dumpSimulationWith simulateStep meta mine moves

