
-- | 2D Radix-8 tree, done naively

{-# LANGUAGE BangPatterns #-}
module Spatial where

--------------------------------------------------------------------------------

import Data.Bits
import Data.List

import PosSet
-- import Data.Array

import Data.Map (Map) ; import qualified Data.Map as Map

--------------------------------------------------------------------------------

newtype BSet = BSet (Map (Int,Int) PSet)

quantize :: Pos -> (Int,Int)
quantize !p = let OldPos x y = unPos' p in (shiftR x 3, shiftR y 3)

--------------------------------------------------------------------------------

bsetInsert :: Pos -> BSet -> BSet
bsetInsert !pos (BSet !bset) = BSet (Map.alter f (quantize pos) bset) where
  f Nothing     = Just $ setSingleton pos
  f (Just pset) = Just $ setInsert    pos pset
  
bsetDelete :: Pos -> BSet -> BSet
bsetDelete !pos (BSet !bset) = BSet (Map.alter f (quantize pos) bset) where
  f Nothing     = Nothing
  f (Just pset) = Just $ setDelete pos pset  

bsetEmpty :: BSet
bsetEmpty = BSet Map.empty

bsetSingleton :: Pos -> BSet
bsetSingleton pos = BSet (Map.singleton (quantize pos) (setSingleton pos))

bsetFromList :: [Pos] -> BSet
bsetFromList = bsetUnion1 bsetEmpty 

bsetToPset :: BSet -> PSet
bsetToPset (BSet bset) = foldl' setUnion setEmpty (Map.elems bset)

bsetToUnsortedList :: BSet -> [Pos]
bsetToUnsortedList (BSet bset) = concat $ map setToList $ Map.elems bset

bsetUnion1 :: BSet -> [Pos] -> BSet   
bsetUnion1 bset stuff = foldr bsetInsert bset stuff

bsetDifference1 :: BSet -> [Pos] -> BSet   
bsetDifference1 bset stuff = foldr bsetDelete bset stuff

--------------------------------------------------------------------------------

nonemptyBSetFindClosest :: Int -> Int -> Pos -> BSet -> [Pos]
nonemptyBSetFindClosest iniradius maxradius center bset = go iniradius where
  go radius = case bsetFindClose radius center bset of 
    []    -> if radius < maxradius then go (radius+10) else []
    stuff -> stuff
    
-- teglatest alakban keresunk, de leszarom
bsetFindClose :: Int -> Pos -> BSet -> [Pos]
bsetFindClose radius center (BSet bset) = concat stuff where
  OldPos cx cy = unPos' center
  x1 = shiftR (cx - radius) 3
  x2 = shiftR (cx + radius) 3
  y1 = shiftR (cy - radius) 3
  y2 = shiftR (cy + radius) 3
  stuff = [ maybe [] setToList (Map.lookup (qx,qy) bset) | qx<-[x1..x2] , qy<-[y1..y2] ]
    
--------------------------------------------------------------------------------
    