
module Main where

--------------------------------------------------------------------------------

import Control.Monad

import Data.IORef
import System.IO
import System.Environment

import System.Posix.Signals
import System.Exit

import Ansi
import Mine
import Simulation
import SimBetter
import Search
import AStar
import Planner

--------------------------------------------------------------------------------

main = do
  -- main_submission
  main_normal

--------------------------------------------------------------------------------
  
main_normal = do
  writeIORef submissionMode False 
  args <- getArgs
  case args of
    [fn]         -> main_sim fn
    ["sim"    ,fn] -> main_sim fn
    ["solve"  ,fn] -> main_strat 0 fn
    ["solsim" ,fn] -> main_strat 1 fn
    ["soldump",fn] -> main_strat 2 fn
    ["play"   ,fn,moves] -> main_play fn moves
    ["dump"   ,fn,moves] -> main_dump fn moves
    ["astar"  ,fn,x,y]   -> main_astar_debug fn (read x) (read y) 
    
    _ -> error "usage: main [sim|solve|solsim|soldump|play|dump] <filepath> {moves}"
 
--------------------------------------------------------------------------------

-- we just abort, who cares :)
my_sigINT_handler :: IO ()
my_sigINT_handler = do
  putChar 'A'
  -- exitWith ExitSuccess
  raiseSignal 9

main_submission = do 
  hSetBuffering stdin  NoBuffering
  hSetBuffering stdout NoBuffering
  hSetEcho stdin False
  
  installHandler sigINT (Catch my_sigINT_handler) Nothing
  
  (meta,mine) <- stdinMap 
  moves <- the_strategy meta mine
  return ()  
  
--------------------------------------------------------------------------------
      
main_strat interactive fn = do
  (meta,mine) <- loadMap fn -- "../maps/contest1.map"
  moves <- the_strategy meta mine
  saveHistory "_last.strategy" (History moves)
  case interactive of
    1 -> do
           putStrLn "=== press space! ==="
           waitForSpace
           timeMachine meta mine moves
    2 -> do 
           dumpSimulationWith simulateStep meta mine moves
    _ -> return ()
    
 --------------------------------------------------------------------------------
  